package mixer;

import java.util.ArrayList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.Random;

/**
 * Created by Tzlil on 12/15/2015.
 */
public class MixNetwork {

    private final Switch[][] switches;

    public MixNetwork(RandomPermutation randomPermutation)
    {
        int[] permutation = randomPermutation.permutation;
        int n = permutation.length;
        assert ((n & n-1) == 0); //n == 2^k
        int layers = (int)(2*Math.log(n)/Math.log(2)) - 1;

        int[] pi, piL, piR;
        Queue<int[]> permutationsQueue = new ArrayBlockingQueue<int[]>(n);
        Graph graph;
        int iDiv2;
        int nDiv2 = n >> 1;
        switches = new Switch[layers][nDiv2];
        int index1,index2;

        permutationsQueue.add(permutation);
        for (int i = n, layer = 0; i > 1; i >>= 1, layer++) // i == permutation size
        {
            iDiv2 = i >> 1;
            for (int j = 0; j < nDiv2; j += iDiv2) // j == permutation start index
            {
                pi = permutationsQueue.remove();
                graph = new Graph(pi);
                piL = new int[iDiv2];
                piR = new int[iDiv2];
                for (int k = 0; k < iDiv2; k++) // k == switch index in permutation j
                {
                    index1 = k + (j << 1);
                    index2 = index1 + iDiv2;
                    switches[layers - layer - 1][k + j] = new Switch(index1,index2,layers - layer - 1,graph.getSwitchValue(k, true));
                    switches[layer][k + j] = new Switch(index1,index2,layer,graph.getSwitchValue(k, false));

                    if (!switches[layers - layer - 1][k + j].value) {
                        piL[k] = pi[k] % iDiv2;
                        piR[k] = pi[k + iDiv2] % iDiv2;
                    } else {
                        piL[k] = pi[k + iDiv2] % iDiv2;
                        piR[k] = pi[k] % iDiv2;
                    }
                }
                permutationsQueue.add(piL);
                permutationsQueue.add(piR);
            }
        }
    }

    public Switch[] getSwitchesByLayer(int layer)
    {
        return switches[layer];
    }

    private class Graph {
        private int n;
        private int nDiv2;
        private Node[][] nodes;
        protected Graph(int[] permutation){
            n = permutation.length; // n = 2^k
            nDiv2 = n >> 1;
            createNodes();
            createEdges(permutation);
            setSwitches();
        }

        // provide an access to algorithm result
        // index must be less then n/2
        protected boolean getSwitchValue(int index,boolean up)
        {
            return up ? nodes[0][index].value : nodes[1][index].value;
        }

        // create two lines of nodes size n/2 each
        // the value of the i th node is (i,i+n/2) if i < n /2  (first line)
        // otherwise its value is (i - n/2 , i)                 (second line)
        private void createNodes()
        {
            nodes = new Node[2][nDiv2];
            for (int i = 0; i < nDiv2; i++)
            {
                nodes[0][i] = new Node();
                nodes[1][i] = new Node();
            }
        }

        // create an edge between each pair of nodes i,j from different lines (i index of the first line)
        // if exists k in  i th node's value and t in j th node's value
        // s.t permutation[k] == t
        // the edge is broken if (k < n/2 and t >= n/2) or (k >= n/2 and t < n/2)
        // Note: in purpose to avoid edge cases, each node has exactly two edges
        private void createEdges(int[] permutation)
        {
            int j;
            for (int i = 0; i < nDiv2; i++)
            {
                j = permutation[i] % nDiv2;
                nodes[0][i].edges.add(new Edge(nodes[1][j], (permutation[i] >= nDiv2)));
                nodes[1][j].edges.add(new Edge(nodes[0][i], (permutation[i] >= nDiv2)));

                j = permutation[i + nDiv2] % nDiv2;
                nodes[0][i].edges.add(new Edge(nodes[1][j], (permutation[i + nDiv2] < nDiv2)));
                nodes[1][j].edges.add(new Edge(nodes[0][i], (permutation[i + nDiv2] < nDiv2)));
            }
        }

        // set switch's value (on/off) for each switch (node)
        // s.t if nodes i,j connected by edge e, i th switch's value
        // must be equal to j's if e is broken or not equal if e is not broken
        private void setSwitches()
        {
            Node node;
            boolean v;
            Edge e0,e1;
            // iterate over first line of nodes
            for (int i = 0; i < nDiv2; i++)
            {
                node = nodes[0][i];
                if (node.set)
                    continue;
                //select default value for first node in connected component
                v = false;
                // set value to all reachable nodes from node
                while (true)
                {
                    node.set = true;
                    node.value = v;
                    e0 = node.edges.get(0); e1 = node.edges.get(1);
                    if (e0.neighbor.set && e1.neighbor.set)
                        break;
                    v ^= (!e0.neighbor.set) ? e0.broken : e1.broken;
                    node = (!e0.neighbor.set) ? e0.neighbor : e1.neighbor;
                }
            }
        }

        //inner classes
        private class Node
        {
            public List<Edge> edges;
            private boolean value;
            private boolean set;
            public Node()
            {
                edges = new ArrayList<Edge>(2);
                set = false;
            }
        }

        private class Edge
        {
            public Node neighbor;
            public boolean broken;
            public Edge(Node neighbor, boolean broken)
            {
                this.neighbor = neighbor;
                this.broken = broken;
            }
        }
    }


}
