package mixer;

import java.math.*;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.crypto.mixnet.MixerOutput;
import qilin.primitives.concrete.ECGroup;
import qilin.util.Pair;
import java.util.Random;

import com.google.protobuf.InvalidProtocolBufferException;

import meerkat.protobuf.Crypto.*;
import meerkat.protobuf.Mixing.*;

import meerkat.crypto.Encryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import verifier.Verifier;

public class Mixer implements meerkat.crypto.mixnet.Mixer {

	protected final Random random;
    protected final Mix2ZeroKnowledgeProver prover;
    protected final Encryption encryptor;
    protected final Mix2ZeroKnowledgeVerifier verifier;

	public Mixer(Random rand, Mix2ZeroKnowledgeProver prov, Encryption enc) {
		this.random = rand;
		this.prover = prov;
		this.encryptor = enc;
		this.verifier = null;
	}

    public Mixer(Random rand, Mix2ZeroKnowledgeProver prov, Encryption enc,Mix2ZeroKnowledgeVerifier verifier) {
        this.random = rand;
        this.prover = prov;
        this.encryptor = enc;
        this.verifier = verifier;
    }

    // assert n = 2^k and n > 1
    protected boolean inputBasicCheckSum(int n){
        return  n > 1 && ((n & (n - 1)) == 0);
    }
    protected RerandomizableEncryptedMessage[][] initializeEncryptionTable(int n,int layers,List<RerandomizableEncryptedMessage> ciphertexts){
        // set first level of encryption
        RerandomizableEncryptedMessage[][] encryptionTable = new RerandomizableEncryptedMessage[layers + 1][n];
        for (int j = 0; j < n; j++) {
            encryptionTable[0][j] = ciphertexts.get(j);
        }
        return  encryptionTable;
    }

    protected  EncryptionRandomness[][] generateRandomnessesForRerandomize(int n,int layers){
        EncryptionRandomness[][] randomnesses = new EncryptionRandomness[layers][n];
        for (int layer = 0; layer < layers; layer++)
        {
            for (int i = 0; i < n; i++) {
                randomnesses[layer][i] = encryptor.generateRandomness(random);;
            }
        }
        return randomnesses;
    }

    protected MixNetwork createMixNetwork(int n){
        return new MixNetwork(new RandomPermutation(n,random));
    }

    protected void rerandomize(int layers,MixNetwork mixNetwork, RerandomizableEncryptedMessage[][] encryptionTable
            ,EncryptionRandomness[][] randomnesses) throws InvalidProtocolBufferException {
        Switch[] switchesLayer;
        int index1,index2;
        RerandomizableEncryptedMessage e1,e2;
        EncryptionRandomness r1,r2;
        for (int layer = 0; layer < layers; layer++)
        {
            switchesLayer = mixNetwork.getSwitchesByLayer(layer);
            for (Switch sw : switchesLayer) {
                index1 = sw.i;
                index2 = sw.j;
                e1 = encryptionTable[layer][index1];
                e2 = encryptionTable[layer][index2];

                r1 = randomnesses[layer][index1];
                r2 = randomnesses[layer][index2];
                if (!sw.value) {
                    encryptionTable[layer + 1][index1] = encryptor.rerandomize(e1, r1);
                    encryptionTable[layer + 1][index2] = encryptor.rerandomize(e2, r2);

                } else {
                    encryptionTable[layer + 1][index1] = encryptor.rerandomize(e2, r2);
                    encryptionTable[layer + 1][index2] = encryptor.rerandomize(e1, r1);
                }
            }
        }
    }

    protected ZeroKnowledgeProof[][] createZKPTable(int n,int layers,MixNetwork mixNetwork, RerandomizableEncryptedMessage[][] encryptionTable
            ,EncryptionRandomness[][] randomnesses) throws InvalidProtocolBufferException {
        Switch[] switchesLayer;
        int index1,index2;
        int switchIndex = 0;
        int nDiv2 = n >> 1;
        ZeroKnowledgeProof[][] proofsTable = new ZeroKnowledgeProof[layers][nDiv2];


        RerandomizableEncryptedMessage e1,e2;
        EncryptionRandomness r1,r2;
        for (int layer = 0; layer < layers; layer++)
        {
            switchesLayer = mixNetwork.getSwitchesByLayer(layer);
            for (Switch sw : switchesLayer) {
                index1 = sw.i;
                index2 = sw.j;
                e1 = encryptionTable[layer][index1];
                e2 = encryptionTable[layer][index2];
                r1 = randomnesses[layer][index1];
                r2 = randomnesses[layer][index2];

                proofsTable[layer][switchIndex] =
                        prover.prove(e1, e2, encryptionTable[layer + 1][index1],
                                encryptionTable[layer + 1][index2],
                                sw.value, sw.i, sw.j, sw.layer, r1, r2);

                switchIndex = (switchIndex + 1) % nDiv2;
            }
        }
        return proofsTable;
    }

	public MixerOutput mix(List<RerandomizableEncryptedMessage> ciphertexts) throws InvalidProtocolBufferException {

		int n = ciphertexts.size();

		if (!inputBasicCheckSum(n))
			return null;

		int layers = (int) (2 * Math.log(n) / Math.log(2)) - 1;  // layers = 2logn -1
		RerandomizableEncryptedMessage[][] encryptionTable = initializeEncryptionTable(n,layers,ciphertexts);
        EncryptionRandomness[][] randomnesses = generateRandomnessesForRerandomize(n,layers);
        MixNetwork mixNetwork = createMixNetwork(n);

        BigInteger time = BigInteger.valueOf(System.currentTimeMillis());
        rerandomize(layers,mixNetwork,encryptionTable,randomnesses);
        System.out.println("re enc time : " + BigInteger.valueOf(System.currentTimeMillis()).subtract(time));

        time = BigInteger.valueOf(System.currentTimeMillis());
        ZeroKnowledgeProof[][] proofsTable = createZKPTable(n,layers,mixNetwork,encryptionTable,randomnesses);
        System.out.println("zkp time : " + BigInteger.valueOf(System.currentTimeMillis()).subtract(time));

		return new mixer.MixerOutput(n,layers,proofsTable, encryptionTable);
	}

}