package verifier;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import prover.ElGamalProofOrganizer;
import qilin.primitives.RandomOracle;
import qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;


public class Verifier implements Mix2ZeroKnowledgeVerifier {


    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final ECElGamalEncryption encryptor;
    private final ECPoint g,h;
    private final ElGamalProofOrganizer organizer;
    private final ZeroKnowledgeOrProofParser parser;

    public Verifier(ECElGamalEncryption encryptor, RandomOracle randomOracle) {
        this.encryptor = encryptor;
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.randomOracle = randomOracle;
        this.organizer = new ElGamalProofOrganizer(group,g,h);
        this.parser = new ZeroKnowledgeOrProofParser(group);
    }

    public BigInteger hash(Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle input) {
        byte[] arr = input.toByteArray();
        return new BigInteger(1,this.randomOracle.hash(arr,arr.length));
    }
    /**
     * @return  true iff the proof is valid
     */
    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.ZeroKnowledgeProof proof) throws InvalidProtocolBufferException {
        ElGamalProofOrganizer.ElGamalProofInput input = organizer.createProofInput(in1,in2,out1,out2);
        return verifyElGamaOrProof(input.getOrProofInput(ElGamalProofOrganizer.OrProofOrder.first), proof.getFirst())&&
                verifyElGamaOrProof(input.getOrProofInput(ElGamalProofOrganizer.OrProofOrder.second), proof.getSecond())&&
                verifyElGamaOrProof(input.getOrProofInput(ElGamalProofOrganizer.OrProofOrder.third), proof.getThird())&&
                verifyElGamaOrProof(input.getOrProofInput(ElGamalProofOrganizer.OrProofOrder.fourth), proof.getFourth());

    }



    public boolean verifyElGamaOrProof(ElGamalProofOrganizer.OrProofInput orProofInput,
                                       Mixing.ZeroKnowledgeProof.OrProof orProof) {

        ZeroKnowledgeOrProofParser.ZeroKnowledgeOrProofContainer container = parser.parseOrProof(orProof);
        List<Condition> conditions = createConditionsList(orProofInput,container);
        boolean result = true;
        for (Condition condition: conditions) { //temporary
            if(!condition.test()){
                if (result){
                    result = false;
                    System.out.print("\n\n\n");
                }
                System.out.println(condition.errorDescription);
            }
        }
        return result;
    }


    private List<Condition> createConditionsList(ElGamalProofOrganizer.OrProofInput orProofInput
            ,ZeroKnowledgeOrProofParser.ZeroKnowledgeOrProofContainer container){
        List<Condition> conditions = new ArrayList<Condition>();
        conditions.add(new Condition<ECPoint>( container.g1,orProofInput.g1,"g1 != g"));
        conditions.add(new Condition<ECPoint>( container.h1,orProofInput.h1,"h1 != e1New.c1/e1.c1"));
        conditions.add(new Condition<ECPoint>( container.g2,orProofInput.g2,"g2 != h"));
        conditions.add(new Condition<ECPoint>( container.h2,orProofInput.h2,"h2 != e1New.c2/e1.c2"));
        conditions.add(new Condition<ECPoint>( container.g1Tag,orProofInput.g1Tag,"g1Tag != g"));
        conditions.add(new Condition<ECPoint>( container.h1Tag,orProofInput.h1Tag,"h1Tag != e2New.c1/e2.c1"));
        conditions.add(new Condition<ECPoint>( container.g2Tag,orProofInput.g2Tag,"g2Tag != h"));
        conditions.add(new Condition<ECPoint>( container.h2Tag,orProofInput.h2Tag,"h2Tag != e2New.c2/e2.c2"));
        conditions.add(new Condition<BigInteger>(container.c1.add(container.c2).mod(group.orderUpperBound()),
                hash(container.forRandomOracle).mod(group.orderUpperBound()).mod(group.orderUpperBound()),
                "(c1 + c2 ) % group size == hash (imput + step1) % group size"));
        conditions.add(new Condition<ECPoint>( group.multiply(container.g1, container.z),
                group.add(container.u, group.multiply(container.h1,container.c1)),"g1 ^ z != u * ( h1 ^ c1 )"));
        conditions.add(new Condition<ECPoint>( group.multiply(container.g2, container.z),
                group.add(container.v, group.multiply(container.h2,container.c1)),"g2 ^ z != v * ( h2 ^ c1 )"));
        conditions.add(new Condition<ECPoint>( group.multiply(container.g1Tag, container.zTag),
                group.add(container.uTag, group.multiply(container.h1Tag,container.c2))
                ,"g1Tag ^ zTag != uTag * ( h1Tag ^ c2 )"));
        conditions.add(new Condition<ECPoint>( group.multiply(container.g2Tag, container.zTag),
                group.add(container.vTag, group.multiply(container.h2Tag,container.c2))
                ,"g2Tag ^ zTag != vTag * ( h2Tag ^ c2 )"));
        return conditions;
    }

    private class Condition<T>{

        T given, expected;
        String errorDescription;

        Condition(T given,T expected,String description){
            this.given = given;
            this.errorDescription = description;
            this.expected = expected;
        }

        boolean test(){
            return given.equals(expected);
        }
    }
}
